"""
synapse-diaspora-auth: A diaspora authenticator for matrix synapse.
Copyright (C) 2017-2018 Shamil K Muhammed.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections.abc import Awaitable, Callable
from typing import Any

import asyncio

import synapse
from synapse import module_api

import bcrypt

import logging


__VERSION__ = "0.2.2"

logger = logging.getLogger(__name__)


class DiasporaAuthProvider:
    __version__ = __VERSION__

    def __init__(self, config: dict, api: module_api):
        self.config = config
        self.api = api
        if self.config.engine == "mysql":
            import aiomysql

            self.module = aiomysql
        elif self.config.engine == "postgres":
            import aiopg

            self.module = aiopg

        api.register_password_auth_provider_callbacks(
            auth_checkers={
                ("m.login.password", ("password",)): self.check_pass,
            },
        )

    async def exec_query(
        self,
        loop: asyncio.AbstractEventLoop,
        query: str,
        *args: list[str],
    ) -> tuple[tuple[str]] | None:
        pool = await self.module.create_pool(
            db=self.config.db_name,
            user=self.config.db_username,
            password=self.config.db_password,
            host=self.config.db_host,
            port=self.config.db_port,
            loop=loop,
        )
        try:
            async with pool.acquire() as connection:
                async with connection.cursor() as cursor:
                    await cursor.execute(query, args)
                    results = cursor.fetchall()
            return results
        except self.module.Error as e:
            logger.warning("Error during execution of query: {}: {}".format(query, e))
            return None
        finally:
            pool.close()
            await pool.wait_closed()

    async def check_pass(
        self,
        username: str,
        login_type: str,
        login_dict: "synapse.module_api.JsonDict",
    ) -> (
        tuple[
            str,
            Callable[["synapse.module_api.LoginResponse"], Awaitable[None]] | None,
        ]
        | None
    ):
        if login_type != "m.login.password":
            return None

        password = login_dict.get("password")
        if not password:
            return None

        local_part = username.split(":", 1)[0][1:] if ":" in username else username
        loop = asyncio.get_event_loop()
        users = await loop.run_until_complete(
            self.exec_query(
                loop,
                "SELECT username, encrypted_password, email FROM users WHERE username=%s",
                local_part,
            )
        )
        # user_id is @localpart:hs_bare. we only need the localpart.
        logger.info("Checking if user {} exists.".format(local_part))

        # check if the user exists.
        if not users:
            logger.info(
                "User {} does not exist. Rejecting auth request".format(local_part)
            )
            return None

        logger.debug("User {} exists. Checking password".format(local_part))
        # user exists, check if the password is correct.
        user = users[0]
        encrypted_password = user[1]
        email = user[2]
        peppered_pass = "{}{}".format(password, self.config.pepper)
        if not (
            bcrypt.hashpw(
                peppered_pass.encode("utf8"), encrypted_password.encode("utf8")
            )
            == encrypted_password.encode("utf8")
        ):
            logger.info(
                "Password given for {} is wrong. Rejecting auth request.".format(
                    local_part
                )
            )
            return None
        return (self.api.get_qualified_user_id(username), None)

    @staticmethod
    def parse_config(config):
        class _Conf:
            pass

        Conf = _Conf()
        Conf.engine = config["database"]["engine"]
        Conf.db_name = (
            "diaspora_production"
            if not config["database"]["name"]
            else config["database"]["name"]
        )
        Conf.db_host = config["database"]["host"]
        Conf.db_port = config["database"]["port"]
        Conf.db_username = config["database"]["username"]
        Conf.db_password = config["database"]["password"]
        Conf.pepper = config["pepper"]
        return Conf
